/*!
  \file h/Loggers.h

  \brief Logging facilities
  
  $Id: Loggers.h,v 1.32 2006/06/12 14:34:53 vit Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   25.07.2005 13:58:39 MSD
*/

#ifndef SK_LOGGERS_H
#define SK_LOGGERS_H

#include "Cfg.h"
#include "utils.h"

#include <loki/Functor.h>
#include <loki/Singleton.h>
#include <loki/AssocVector.h>

#include <iostream>
#include <list>

extern "C" {
#include <stdarg.h>
#include <errno.h>
};

/// This enumeration represents different widely-used log levels
typedef enum {
	/// Debug message
	LlDbg = 0,
	/// Informational message
	LlInfo,
	/// Warning message
	LlWarning,
	/// Error message
	LlError,
	/// Critical error message
	LlCritical
} LogLevel_t;		

/*! \brief Logging function.
  
    Pointer to this function is an optional parameter of any ++Skype class constructor with logging ability.
    This functor has 3 parameters: log level, format string and variable list of arguments To know more about va_list 
    see vprintf description, for example.
*/    
typedef Loki::Functor<void, LOKI_TYPELIST_3(const LogLevel_t, const char*, va_list)> LogFunc_t;

/// Dummy (pseudo) logging device. This device is default logging device for ++Skype classes 
inline void DummyLogger(const LogLevel_t _lvl, const char* _fmt, va_list _args) 
{
};

/// This class grants logging ability to it's derived classes.  The core logging function is called log
class LogAbility {
public:
	/*!
  	  \brief Stores the reference to callback functor used for real logging

  	  \param _logger Call-back functor used by log method
	*/
	LogAbility(const LogFunc_t& _logger = DummyLogger) : logger(_logger) {};
	/*!
	  \brief Simple log function. It looks (almost) like printf. 

	  \param _lvl Log level

	  \param _fmt Log string format. Its specifications depends on real logger used but usually it equals to 
	         printf format specification.
	*/  
	void log(LogLevel_t _lvl, const char* _fmt, ...) const;
	/// Returns logger functor associated with the object
	const LogFunc_t& GetLog() const;
	/*!
	  \brief Set new logger associated with the object

	  \param _logger New logger function
	*/
	void SetLog(const LogFunc_t& _logger);
private:
	/// Logger functor
	mutable LogFunc_t logger;
};

/// Simple template to know the default minimum log-level for different compilation modes
template <DbgMode_t DBG = DbgMode> class  DefaultMinLogLevel {
public:
	/// Default log level in debug mode is LlDbg
	static LogLevel_t GetLvl() { return LlDbg; }; 	
};

/// DefaultMinLogLevel specialization for release mode of compilation
template <> class DefaultMinLogLevel<release> {
public:
	/// Default log level in release mode is LlInfo
	static LogLevel_t GetLvl() { return LlInfo; }; 	
};

/// Log prefix policy - no prefix (log messages are not prefixed by anything)
class NoLogPrefix {
public:
	/// Returns empty string (empty prefix)
	std::string& operator()(const LogLevel_t& _lvl);
private:
	/// Empty strings
	std::string empty;
};

/// Log prefix policy - log level prefix (log messages are prefixed by log level)
class LvlLogPrefix {
public:
	/// Initialization
	LvlLogPrefix();	
	/// Returns textual representation of the log level
	std::string& operator()(const LogLevel_t& _lvl);
private:
	/// Buffer to store the textual representation of the log level
	std::string buf;
	/// Maps log levels to their textual representation 
	Loki::AssocVector<LogLevel_t, std::string> levels_str;
};

/// Log prefix policy - log level & time prefix (log messages are prefixed by log level and time of the message) 
class LvlTimeLogPrefix : public LvlLogPrefix {
public:
	/// Returns textual representation of the log level and current date/time
	std::string& operator()(const LogLevel_t& _lvl); 
};

/*!
  \brief Base (and trivial) logger functionality

  This template is used to simplify loggers declarations and implementations. The main logic it implements is
  the "minimal log level" logic.  This template can not be used alone.  It should be used only with host class.
  Host's void Do(const char* _fmt, va_list _args) will be called by 
  BaseLogger::operator()(LogLevel_t _lvl, const char* _fmt, va_list _args) if the _lvl argument is greater or equal than 
  minimal log level.	

  Loggers are frequently used as singletons in the modern world of software design.  That's why logger's constructors
  can be called before entering main() function. So, it would not be a perfect design style to throw exceptions in logger's
  constructors. On the other hand, there is a common solution to initialize logger resources inside its constructors.  It is
  clear, that such initialization procedure can fail. It is necessary to know the result of initialization after entering 
  main().  #IsReady/#SetUnReady methods pair of this class seems to be the solution of this dilemma.  #IsReady is public 
  method which should be used to check loggers's "ready for the mission" (successfull initialization) status.
  #SetUnReady is protected method which should be called by derived class's constructors instead of exception throwing.
*/  
template <class Host, class Prefix = NoLogPrefix> class BaseLogger {
public:
	/*!
  	  \brief  Stores reference to host (to make possible host.OutMsg call in operator()) and
	  	  minimal log level. Messages with log level less than minimum log level will not be logged.

  	  \param  _host		Reference to host object	
  	  \param  _min_lvl	Minimal log level
	*/
	BaseLogger(Host& _host, const LogLevel_t _min_lvl = DefaultMinLogLevel<>::GetLvl());
	/*!
  	  \brief Sets minimal log level. Messages with log level less than minimum log level will not be logged.
  
  	  \param _min_lvl New value of minimal log level.
	*/
	void SetMinLogLvl(const LogLevel_t _min_lvl) { min_lvl = _min_lvl; };
	/*!
  	  \brief  Log function. It is compatible with LogFunc_t declaration.  That's why obejcts of derived classes
	  	  can be used as loggers.

		  Checks _lvl and if it is greater or equal then minumal log level, calls host's Do function. 

  	  \param _lvl  Log-level. Messages with log level less than minimum log level will not be logged
	  \param _fmt  Format of log string.  See (v)printf manual for details.  
	  \param _args Arguments related to format.  See vprintf manual for additional information.
	*/
	void operator()(const LogLevel_t _lvl, const char* _fmt, va_list _args);
	/*!
	  \brief Is this logger ready for logging?
  
	  \param _err Reference to string where error message should be placed in case where logger is not ready.
	  
	  \return true if logger is ready for logging, otherwise returns false and sets _err variable to human-readable 
	          error message. 
	*/
	bool IsReady(std::string& _err);
protected:
	/*!
  	  \brief Set "ready for the logging" status to false
		 
	  	 This method should be called by host whenever critical error with loging device occurs.
		 
  	  \param _err Text error message.
	*/
	void SetUnReady(const std::string& _err);
	/*!
  	  \brief Set "ready for the logging" status to true
		 
	*/
	void SetReady(void);
private:
	/// Reference to host object
	Host& host;
	/// Prefix generator
	Prefix prefix;
	/// Minimal log level.
	LogLevel_t min_lvl;
	/// Are we ready for the mission (logging)?
	bool is_ready;
	/// Error message if is_ready equals to false
	std::string err_msg;
	/// Bufer used to write log-level text presentation into output log device
	std::string buf;
};

/*!
  \brief Logging device - file

  This class can be used for logging into files. See also documentation on BaseLogger for additional useful remarks.
*/  
template <class Prefix = LvlLogPrefix>
class FileLogger : public BaseLogger<FileLogger<Prefix>, Prefix> {
public:
	/*!
  	  \brief  Opens log file in append mode.  Any error can be checked via IsReady method.

  	  \param _name 		The name of log file.
	  \param _flush_period  Flush period (in lines) 
	  \param _min_lvl 	Minimal log level.  Messages with log level less than _min_lvl will not be logged
	*/
	FileLogger(const std::string& _name, const unsigned _flush_period = 10, 
								const LogLevel_t _min_lvl = DefaultMinLogLevel<>::GetLvl());
	/*!
  	  \brief This constructor is for logging into already open files.  stderr and stdout files are typical
	  	 examples.     
  	  
	  \param _f		Pointer to the open file. It should be writable, of course.
	  \param _flush_period  Flush period (in lines) 
	  \param _min_lvl	Minimal log level.  Messages with log level less than _min_lvl will not be logged.
	  \param _owner 	File closing policy.  If _owner is false file _f will not be closed in FileLogger's destructor,
	         		else will be.
	*/
	FileLogger(FILE* _f, const unsigned _flush_period = 10, const LogLevel_t _min_lvl = DefaultMinLogLevel<>::GetLvl(), 
													bool _owner = false);
	/*!
  	  \brief  Real logging function.  See BaseLogger documentation for additional information.
		
	  	  Flushes the output every period (see documentation on FileLogger constructors for additional details)	  
	  
	  \param _lvl  Log-level. Messages with log level less than minimum log level will not be logged
	  \param _fmt  Format of log string.  See (v)printf manual for details.  
	  \param _args Arguments related to format.  See vprintf manual for additional information.
	*/
	void Do(const LogLevel_t _lvl, const char* _fmt, va_list _args);
	/// Closes file if this object is it's owner. 
	~FileLogger();
private:
	/// Don not copy an objects of this class
	FileLogger(const FileLogger& _logger);
	/// Don not assign an objects of this class
	void operator=(const FileLogger& _logger);
	std::string name;
	/// File handle
	FILE* f;
	/// Should I close file in destructor?
	bool should_close;
	/// Counter of "Do" calls
	unsigned do_cntr;
	/// Flus period in lines
	const unsigned flush_period;
};

/*!
  \brief  Logging device - stream 

  This class can be used for logging into streams. See also documentation on BaseLogger for additional useful remarks.
*/  
template <class Prefix = LvlLogPrefix>
class StreamLogger : public BaseLogger<StreamLogger<Prefix>, Prefix> {
public:
	/*!
  	  \brief Construct logger from output stream.

  	  \param _stream  Reference to output stream opened foe writing.
  	  \param _min_lvl Minimal log level.  Messages with log level less than _min_lvl will not be logged.
	*/
	StreamLogger(std::ostream& _stream = std::cerr, const LogLevel_t _min_lvl = DefaultMinLogLevel<>::GetLvl());
	/*!
  	  \brief  Real logging function.  See BaseLogger documentation for additional information.
   	
	  \param _lvl  Log-level. Messages with log level less than minimum log level will not be logged
	  \param _fmt   Format of log string.  See (v)printf manual for details.  
	  \param _args  Arguments related to format.  See vprintf manual for additional information.
	*/
	void Do(const LogLevel_t _lvl, const char* _fmt, va_list _args);
private:	
	/// Output stream
	std::ostream& stream;
	/// Bufer used to format output string before writing it into stream
	char str[512];
};


/*!
  \brief  Chain of loggers. Forward log message to all loggers, which was added to chain

  Usually used as singletons for logging from the application
*/  
class LoggersChain : public LogAbility {
public:
  	/// Constructs the chain of loggers.
	LoggersChain() throw();
	/*! 
  	  \brief Add logger (one of logger devices - FileLogger or StreamLogger) to the chain.

	  \param _logger Logger functor.
	 */
	template <typename L> void AddLogger(L& _logger);
	/*! 
  	  \brief Add logger (logger functor) to the chain.

	  \param _logger Logger functor.
	 */
	void AddLogFunc(const LogFunc_t& _logger);
	///  Remove (last added) logger from the chain.
	void PopBackLogger(void);
	/*!
	  \brief  Does real logging - calls all loggers in the chain
  
	  \param _lvl	Level of the log message
	  \param _fmt	Format of the log message
	  \param _args  Log mesage itself
	*/
	void Do(const LogLevel_t _lvl, const char* _fmt, va_list _args);
private:
	/// Copy constructor is prohibited
	LoggersChain(const LoggersChain&);
	/// Assign operator is prohibited
	LoggersChain& operator=(const LoggersChain&);
	/// Chain of loggers
	std::list<LogFunc_t> logger;
};


#include "unix/Loggers.h"

/**********************************************************************************************************/
/************************************ Implementation ******************************************************/
/**********************************************************************************************************/

inline void LogAbility::log(LogLevel_t _lvl, const char* _fmt, ...) const
{
  va_list args;	
  va_start(args, _fmt);
  logger(_lvl, _fmt, args);
  va_end(args);	
};

inline const LogFunc_t& LogAbility::GetLog() const
{ 
  return logger;
};

inline void LogAbility::SetLog(const LogFunc_t& _logger)
{
  logger = _logger;	
};

template <class Host, class Prefix> 
BaseLogger<Host, Prefix>::BaseLogger(Host& _host, const LogLevel_t _min_lvl)
  	         : host(_host), min_lvl(_min_lvl), is_ready(true) 
{};

template <class Host, class Prefix> 
inline void BaseLogger<Host, Prefix>::operator()(const LogLevel_t _lvl, const char* _fmt, va_list _args)
{
  //const char* n = __FUNCTION__;
  
  if(_lvl >= min_lvl) {
	Assert(is_ready);
	std::string& _out = prefix(_lvl);
	//int tst = _out.length();
	if(_out.length()) _out.append(_fmt);
	const std::string& out = (_out.length()) ? _out : _fmt;
	host.Do(_lvl, out.c_str(), _args);

  }; 
};

template <class Host, class Prefix>
inline bool BaseLogger<Host, Prefix>::IsReady(std::string& _err)
{
  if(!is_ready) _err = err_msg;
  return is_ready;
};

template <class Host, class Prefix>
void BaseLogger<Host, Prefix>::SetUnReady(const std::string& _err)
{
  is_ready = false;
  err_msg = _err;
};

template <class Host, class Prefix>
void BaseLogger<Host, Prefix>::SetReady(void)
{
  is_ready = true;
  err_msg = "";
};

template <class Prefix>
FileLogger<Prefix>::FileLogger(const std::string& _name, const unsigned _flush_period, const LogLevel_t _min_lvl)
	   : BaseLogger< FileLogger<Prefix>, Prefix >(*this, _min_lvl), name(_name), should_close(true), flush_period(_flush_period)
{
  f = fopen(_name.c_str(), "a" );
  if(f == NULL) {
	should_close = false;
	BaseLogger< FileLogger<Prefix>, Prefix >::SetUnReady(strerror(errno));
  }	  
};

template <class Prefix>
FileLogger<Prefix>::FileLogger(FILE* _f, const unsigned _flush_period, const LogLevel_t _min_lvl, bool _owner)
	  : name(""), f(_f), should_close(_owner), flush_period(_flush_period),
	    BaseLogger< FileLogger<Prefix>, Prefix >(*this, _min_lvl)
{};

template <class Prefix>
FileLogger<Prefix>::~FileLogger()
{
  if(should_close) fclose(f);	
};

template <class Prefix>
inline void FileLogger<Prefix>::Do(const LogLevel_t _lvl, const char* _fmt, va_list _args)
{
  vfprintf(f, _fmt, _args);
  fprintf(f, "\n");
  if(++do_cntr > flush_period) {
	do_cntr = 0;
	fflush(f);
  };	  
};

template <class Prefix>
inline void StreamLogger<Prefix>::Do(const LogLevel_t _lvl, const char* _fmt, va_list _args)
{
  vsnprintf(str, sizeof(str), _fmt, _args); 
  stream << str << std::endl;
};	

template <class Prefix>
StreamLogger<Prefix>::StreamLogger(std::ostream& _stream, const LogLevel_t _min_lvl)
	     : BaseLogger< StreamLogger<Prefix>, Prefix >(*this, _min_lvl), stream(_stream)
{};

template <typename L> void LoggersChain::AddLogger(L& _logger)
{
  LogFunc_t lf(&_logger, &L::operator());
  AddLogFunc(lf);
};	
#endif // SK_LOGGERS_H
