/*!
  \file User.h

  \brief User-related stuff, declarations

  $Id: User.h,v 1.8 2006/05/23 10:22:14 tarle Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   18.08.2005 20:36:11 MSD
*/

#ifndef SK_USER_H
#define SK_USER_H

#include "SkypeObj.h"

namespace Ice {

/// User's attributes 
typedef enum {
	uHandle,
	uFullName,
	uBirthday,
	uSex,
	uLanguage,
	uCountry,
	uProvince,
	uCity,
	uPhoneHome,
	uPhoneOffice,
	uPhoneMobile,
	uHomePage,
	uAbout,
	uHasCallEquipment,
	uIsVideoCapable,
	uBuddyStatus,
	uIsAuthorized,
	uIsBlocked,
	uDisplayName,
	uOnlineStatus,
	uLastOnlineTimeStamp,
	uCanLeaveVM,
	uSpeeddial,
	uReceiveAuthRequest,
	uMoodText,
	uAliases,
	uTimezone,
	uIsCFActive
} UserAttrs_t;

/// User attributes description
typedef AttrsDescription<UserAttrs_t> UserAttrsDesc_t;

/// Describes atribute of Skype object "User"
template <>
class AttrsDescription<UserAttrs_t> : public AttrsDescriptionBase<UserAttrs_t> {
public:
	/// Fills description
	AttrsDescription();
};

// Forward declaration of User object
class User;

/// Watchers for the User object
typedef StateWatchers<User, UserAttrs_t> UserWatchers_t;

/// This class represents User Skype object
class User : public SkypeObject<User, UserAttrs_t, SkypeIdObjCmdFormatter, SkypeObjRegExParser, std::string> {
public:
    /*!
      \brief	Constructs User object with DEFAULT error handler.  
  		  
	  	Default error handler doesn't handle any error at all.

      \param _sc		Reference to Skype Connection
      \param _handle	User identifier
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */
    User(SkypeConnection& _sc, const std::string& _handle, const LogFunc_t& _log_func = DummyLogger); 
     /*!
      \brief	Constructs User object with NON-DEFAULT (user-supplied) error handler.  
  		  
      \param _sc		Reference to Skype Connection
      \param _handle	User identifier
      \param _eh		Error hanlder.  This functor will be called in case of errors.
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */
    User(SkypeConnection& _sc, const std::string& _handle, const ObjErrHandler_t& _eh, const LogFunc_t& _log_func = DummyLogger);
    /*!
      \brief Checks if object is alive or not. 

      \return alive if object is alive, finished if not alive and unknown if object doesn't held "status" attribute		
    */  
    LiveMode_t IsAlive() const;
private:
    /// Object-specific initialization.  Called from constructors 
    void Init();	
};


/// Possible types of users search
typedef enum {
	srFriends,
	srUsers,
	srUsersWaitingMyAuthorization
} UserSearchTypes_t;

/// Description of the different users search types
typedef AttrsDescription<UserSearchTypes_t> UserSearchDesc_t;

/// Describes users search types
template <>
class AttrsDescription<UserSearchTypes_t> : public AttrsDescriptionBase<UserSearchTypes_t> {
public:
	/// Type of attributes of the oobjects returned by search procedure
	typedef UserAttrs_t ObjAttrs_t; 
	/// Type of objects returned by the search procedure
	typedef User ObjType_t;
	/// Fills description
	AttrsDescription();
};


/*!
  \brief User objects manager

  The manager of users.
*/  
class UserManager : public SkypeObjManager<User, UserAttrs_t, UserManager, UserSearchTypes_t, Searcher> {
public:
	UserManager(SkypeConnection& _sc, const LogFunc_t& _log_func = DummyLogger); 
	UserManager(SkypeConnection& _sc, const User::ObjErrHandler_t& _eh, const LogFunc_t& _log_func = DummyLogger); 
	/*!
	  \brief Parses event related to user
  
	  \param _match Parsed regular expression
	  \return user identifier
	*/
	std::string ParseEvent(const boost::cmatch& _match);

private:
	/// Connection to Skype
	SkypeConnection& sc;
	void Init();
	
};

}; // namespace Ice
#endif // SK_CHAT_H
