/*!
  \file unix/LowLevel-X11.h

  \brief Low-level ++Skype API implementation for unix OS. X11 variant

  $Id $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   18.10.2006 08:19:42 MSD
*/

#ifndef UNIX_LOWLEVEL_X11_H
#define UNIX_LOWLEVEL_X11_H

#include "../Cfg.h"

#ifdef USE_X11_INTERFACE 

#include "../SkypeExceptions.h"
#include "../utils.h"

extern "C" {
#include <assert.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
};

#include <vector>

/// Implementation of RawSkypeConnection interface for the linux X11 API
typedef RawSkypeConnection<os_nix> UnixX11RawSkypeConnection_t;

/// Implementation of RawSkypeConnection interface for the unix (currently linux) API
typedef UnixX11RawSkypeConnection_t UnixRawSkypeConnection_t;

/// RawSkypeConnection specialization for unix OS
template <> class RawSkypeConnection<os_nix> : public SkypeCmdFormatter< RawSkypeConnection<os_nix> >,
					       public LogAbility 
{
public:
	/*!
	  \brief Stores functor which will be executed upon messages (Skype -> application) receiving

	  \param _cbk Call-back functor. It will be executed for any message received from Skype
	  \param _app_name Application's name (usually argv[0]).  
	  \param _log_func Logger to use inside RawSkypeConnection 
	*/  
	RawSkypeConnection(const RawNotify_t& _cbk, const std::string& _app_name, LogFunc_t _log_func = DummyLogger);	
	/*!
	 * \brief		Attach to Skype. Do nothing if already attached. 
	   
	   \param _display	X11 Display to connect to. The display has to be opened by the calling application

	  \throw SkypeDisconnect   Error occured during connect.		
	  \throw SkypeAccessDenied Skype user doesn't allow this application to connect to Skype
	*

	*/ 
	void Attach(Display* _display);
	/*!
  	  \brief		Attach to Skype. Do nothing if already attached. 
	  
	  \param _display	X11 Display to open. If empty, the value of DISPLAY environment 
	                        variable is used.  This function initiates a connection to the display.

	  \throw SkypeDisconnect   Error occured during connect.		
	  \throw SkypeAccessDenied Skype user doesn't allow this application to connect to Skype
	*/
	void Attach(const std::string& _display = "");
	/// Returns true if attached to Skype, false - otherwise
	bool IsAttached(void) { return attach_result; };
	/// Detach from Skype. Do nothing if already detached
	void Detach();
	/*!
	  \brief Send mesage to Skype via D-BUS. 

	  \param _cmd Skype command

	  \throw SkypeDisconnect Disconnected from Skype 
	*/  
	void flush(const std::string& _cmd);
	/*!
	  \brief Send mesage to Skype via D-BUS (calls SkypeCmdFormatter::flush())

	  \throw SkypeDisconnect Disconnected from Skype 
	*/  
	void flush();
	/*!
	  \brief	Read (if any) and process messages from Skype (Skype notifications and 
	  		command responses).

	  		This is a very simple method. It checks for new messages in queue and
			runs user-defined call-back functor (see RawSkypeConnection constructor) 
			for each received message. See also Attach docs for some implementation 
			details.
			
	  \throw AssertionViolation  If method is executed when connection to D-BUS bus is not established.
	  			     It is thrown only if debug mode is on (DbgMode is set to debug 
				     in Cfg.h).
			
	  \return	True if at least 1 message was received.  False otherwise.
	*/  
	bool operator()();
	/*!
	 * \brief  Check if X11 event is Skype-related.  Used internally
	 *
	 * \param _ev X11 event
	 *
	 * \returns true if event _ev relates to Skype, false otherwise
	*/ 
	bool IsMyXEvent(const XEvent& _ev) const; 
	/// Frees all the resources
	~RawSkypeConnection();
private:
	/// Internal loop for X11 events
	void ProcessXEvents();
	/// Call-back functor.  
	RawNotify_t cbk;
	/// Application's name
	const std::string app_name;
	/// Is connection to Skype established?
	bool attach_result;
	/// Should we close the connection to X-server or not?
	bool should_close_display;
	/// X11 Display structure
	Display  *display;
	/// Handle of Skype window
	Window skype_win;
	/// Handle of the window which is created by this object (to send commands/receive events to/from Skype) 
	Window my_win;
	/// X11 atom to get Skype events 
	Atom a2, a1;
	/// Cached events (read from Skype)
	std::vector<std::string> cached_replies;
	/// The reply which is currently being reading from the Skype
	std::string cur_reply;
};

/**********************************************************************************************************/
/************************************ Implementation ******************************************************/
/**********************************************************************************************************/


#endif // USE_X11_INTERFACE 
#endif // UNIX_LOWLEVEL_X11_H

