/*!
  \file win/LowLevel.h

  \brief Low-level ++Skype API implementation for Microsoft Windows

  $Id: LowLevel.h,v 1.20 2006/06/15 09:43:29 tarle Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   20.07.2005 08:19:42 MSD
*/

#ifndef WIN_LOWLEVEL_H
#define WIN_LOWLEVEL_H

#include "../Cfg.h"

#if defined(SKYPE_MINGW) || defined(SKYPE_MSC) 

#include <windows.h>

#include "../SkypeExceptions.h"
#include "../utils.h"

/// Skype attach response codes
enum {
	/// Client is successfully attached and API window handle can be found in wParam parameter
	SKYPECONTROLAPI_ATTACH_SUCCESS = 0,

	/// Skype has acknowledged connection request and is waiting for confirmation from the user.
	SKYPECONTROLAPI_ATTACH_PENDING_AUTHORIZATION=1,	

	/// The client is not yet attached and should wait for SKYPECONTROLAPI_ATTACH_SUCCESS message, user has explicitly denied access to client
	SKYPECONTROLAPI_ATTACH_REFUSED=2,								

	// API is not available at the moment. For example, this happens when no user is currently logged in.
	SKYPECONTROLAPI_ATTACH_NOT_AVAILABLE=3,	

	/// Client should wait for SKYPECONTROLAPI_ATTACH_API_AVAILABLE broadcast before making any further connection attempts.
	SKYPECONTROLAPI_ATTACH_API_AVAILABLE=0x8001
};


/// Implementation of RawSkypeConnection interface for the unix (currently linux) API
typedef RawSkypeConnection<os_win> WinRawSkypeConnection_t;

typedef Loki::Functor<LRESULT, LOKI_TYPELIST_4(HWND, UINT, WPARAM, LPARAM)> WindowProc_t;

/// Temporary solution - pointer to instance of WinRawSkypeConnection_t initialized at #WinRawSkypeConnection_t::WinRawSkypeConnection_t
extern void* pthis;

/// RawSkypeConnection specialization for Microsoft Windows
template <> class RawSkypeConnection<os_win> : public SkypeCmdFormatter< RawSkypeConnection<os_win> >,
	public LogAbility 
{
public:
	/*!
	  \brief Stores functor which will be executed upon messages (Skype -> application) receiving

	  \param _cbk Call-back functor. It will be executed for any message received from Skype
	  \param _app_name Application's name (usually argv[0]).  
	*/  
	RawSkypeConnection(const RawNotify_t& _cbk, const std::string& _app_name, LogFunc_t _log_func = DummyLogger);	
	/*!
  	  \brief Attach to Skype. Do nothing if already attached.
  
	  \param _timeout 	Select's timeout value in microseconds. The value of this parameter
	  			doesn't used anywhere if _dont_select is set to true. 

	  \throw SkypeDisconnect Error occured during connect.		
	  \throw SkypeAccessDenied Skype user doesn't allow this application to connect to Skype
	*/
	void Attach(int _timeout = 3000);
	/// Returns true if attached to Skype, false - otherwise
	bool IsAttached(void) { return attach_result; }
	/// Detach from Skype. Do nothing if already detached
	void Detach();
	/// Send mesage to Skype via Windows Message Queue
	void flush(const std::string& cmd);
	/*!
	  \brief Send mesage to Skype via Windows Message Queue (calls SkypeCmdFormatter::flush())

	  \throw SkypeDisconnect Disconnected from Skype 
	*/  
	void flush();
	/*!
	  \brief	Does nothing on Windows platform
			
	  \return	Always returns false
	*/  
	bool operator()();
private:
	/// Call-back functor.  
	RawNotify_t cbk;
	/// Application's name
	const std::string app_name;
	/// Is connection to Skype via Windows Message Queue established?
	bool attach_result;
	/// Is access denied by Skype?
	bool access_denied;
	/// Indicate "Pending authorization" state of connect to Skype API
	bool pending_authorization;
	/// Indicate that process already call Attach() method
	bool isInAttach;
	/// Windows process handler
	HINSTANCE processHandle;
	/// Service window class name. It is used only for interaction with Skype
	char windowClassName [128];
	/// Service window handle. It is used only for interaction with Skype
	HWND mainWindowHandle;
	/// Handle of Skype window.
	HWND skypeWindowHandle;
	/// Identifier of 'SkypeControlAPIAttach' window message
	UINT skypeControlAPIAttachID;
	/// Identifier of 'SkypeControlAPIDiscover' window message
	UINT skypeControlAPIDiscoverID;
	/*!
	  \brief Procedure of #mainWindowHandle window (see WinAPI documentation), it 
	  call the member function #WindowProc for process Skype message

	  \param hWindow Window handle (See WinAPI documentation)
	  \param uiMessage Messsage ID (See WinAPI documentation)
	  \param uiParam wParam (See WinAPI documentation)
	  \param ulParam lParam (See WinAPI documentation)

	  \return	Result of message process (See WinAPI documentation)
	 */
	static LRESULT APIENTRY _WindowProc(
		HWND hWindow, UINT uiMessage, WPARAM uiParam, LPARAM ulParam);

	/// Process of Skype message.
	LRESULT WindowProc(
		HWND hWindow, UINT uiMessage, WPARAM uiParam, LPARAM ulParam);

	/*! 
	  \brief Create class of service window for interaction with Skype.
	  Init #processHandle and #windowClassName members 

	  \return true if window class for SkypeAPI interaction was regostered successful. 
	  false - otherwise
	 */
	bool Initialize_CreateWindowClass(void);

	/*! 
	  \brief Create service window for interaction with Skype
	  Init #mainWindowHandle member

	  \return true if window for SkypeAPI interaction was created successful. 
	  false - otherwise
	 */
	bool Initialize_CreateMainWindow(void);

};

#endif // SKYP_MINGW || SKYPE_MSC
#endif // WIN_LOWLEVEL_H






